// ---------------------------------------------------------------------------------------------------------------------
// Callbacks requested for this integration service. Default includes "deliveryStatus", "status", "response"
// ---------------------------------------------------------------------------------------------------------------------
CALLBACKS = ["deliveryStatus", "status", "response"];

/*
 Called when response object has been received from xMatters
 */
function apia_response(apxml) {
  log.info("Enter - apia_response");
  log.info("Method = [" + apxml.getMethod() + "]");

  var apxmlAsObj = APXML.toResponseJs(apxml);
  apia_callback(apxmlAsObj);

  log.info("Exit - apia_response");
}

function apia_callback(msg) {
  log.info("Enter - apia_callback");

  IALOG.debug("Received message from xMatters:\nIncident: {0}\nEvent Id: {1}\nCallback Type: {2}",
    msg.incident_id, msg.eventidentifier, msg.xmatters_callback_type);

  try {

    switch (msg.xmatters_callback_type) {

      case "response":
        try {
          handleResponse(msg);
        } catch (e) {
          handleException(msg, e);
        }
        break;

      case "status":
        handleEventStatus(msg);
        break;

      case "deliveryStatus":
        handleDeliveryStatus(msg);
        break;
    }

  } catch (e) {
    // convert the exception to a string, if it isn't one already.
    //var exStr = ( typeof e == "object" ) ? JSON.stringify(e) : e;
    var exStr = ( typeof e == "object" ) ? e.toString() : e;
    log.error("apia_callback(" + msg.eventidentifier + ", " + msg.xmatters_callback_type + "): caught Exception - name: [" + e.name + "], message [" + e.message + "]: " + JSON.stringify(msg) + "\n" + exStr);

    throw e;
  }


  log.info("Exit - apia_callback");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleDeliveryStatus
 *
 * The main routine for handling delivery status messages from xMatters
 *
 * If an message is delivered successfully, a comment is added to the worklog of the BMC Remedy Change Incident ticket quoting the user and device that was targeted
 * If an message delivery fails, a comment is added to the worklog of the BMC Remedy Change Incident ticket quoting the user and device that was targeted
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleDeliveryStatus(msg) {
  log.debug("Enter - handleDeliveryStatus");

  if (msg.deliverystatus) {
    var annotation = "";
    var change = newRemedyObject();
    var infrastructureChangeId = getInfrastructureChangeId(msg);
    var changeInterfaceService = new XMCHGChangeInterfaceWS(new WSUtil(), XM_CHG_CHANGE_INTERFACE_WS_URL);

    switch (String(msg.deliverystatus).toLowerCase()) {
      case "delivered":
        annotation = "[Event " + msg.eventidentifier + "] Notification delivered successfully to " + msg.recipient + " | " + msg.device;
        logNotification(annotation);
        break;

      case "failed":
        annotation = "[Event " + msg.eventidentifier + "] Unable to deliver notification to " + msg.recipient;
        logNotification(annotation);
        break;
    }

    if (ANNOTATE_DELIVERY) {
      // Update status in BMC Remedy Change
      change.Infrastructure_Change_ID = infrastructureChangeId;

      setChangeWorkLog(change, true, NOTE_PREFIX + annotation, null);
      changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, change);
    }
  }

  log.debug("Exit - handleDeliveryStatus");

}

function buildAnnotation(annotation, alertStatus, responder, switchOrder) {
  log.debug("Enter - buildAnnotation");

  if (annotation == null || annotation == "") {
    annotation = (switchOrder ? responder + alertStatus : alertStatus + responder);
  } else {
    annotation = annotation.replace('"', "") + " by " + responder;
  }

  log.debug("Exit - buildAnnotation");
  return annotation;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleResponse
 *
 * The main routine for handling user responses
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleResponse(msg) {
  log.debug("Enter - handleResponse");

  var change = newRemedyObject();
  var changeInterfaceService = new XMCHGChangeInterfaceWS(new WSUtil(), XM_CHG_CHANGE_INTERFACE_WS_URL);
  var infrastructureChangeId = getInfrastructureChangeId(msg);
  var requestId = getRequestId(msg);

  var responder = msg.recipient;
  var device = msg.device;
  var annotation = "null".equals(msg.annotation) ? "" : msg.annotation;

  IALOG.debug("{0}: Received response '{1}' from {2}|{3} with comment '{4}'", infrastructureChangeId, msg.response, responder, device, annotation);

  // Update status in BMC Remedy Change
  change.Infrastructure_Change_ID = infrastructureChangeId;


  // Set the annotations
  switch (String(msg.response).toUpperCase()) {
    case RESPONSE_ACTION_APPROVED:
    case RESPONSE_ACTION_APPROVE:
      annotation = buildAnnotation(annotation, CHANGE_REQUEST_APPROVED, responder);
      break;
    case RESPONSE_ACTION_REJECTED:
    case RESPONSE_ACTION_REJECT:
      annotation = buildAnnotation(annotation, CHANGE_REQUEST_REJECTED, responder);
      break;
    case RESPONSE_ACTION_HOLD:
      annotation = buildAnnotation(annotation, CHANGE_REQUEST_HOLD, responder);
      break;
    case RESPONSE_ACTION_ANNOTATE:
      annotation = buildAnnotation(annotation, CHANGE_REQUEST_ANNOTATE, responder);
      break;
    case RESPONSE_ACTION_VIEW_CHANGE:
      annotation = buildAnnotation(annotation, CHANGE_REQUEST_VIEW_CHANGE, responder, true);
      break;
  }

  // Approve/Reject/Hold the request
  switch (String(msg.response).toUpperCase()) {
    case RESPONSE_ACTION_APPROVED:
    case RESPONSE_ACTION_APPROVE:
    case RESPONSE_ACTION_REJECTED:
    case RESPONSE_ACTION_REJECT:
    case RESPONSE_ACTION_HOLD:

      // Construct a qualification that will return the approval request that the user is responding to.
      // Retrieve the full description of the approval that the event is associated with, to get the Change Request
      var changeAPDetailSignatureService = new XMCHGChangeAPDetailSignatureWS(new WSUtil(), XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_URL);

      var qualification = XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_GET_APPROVAL_BY_ID_QUALIFICATION.replace("${REQUEST_ID}", requestId);
      var approvalList = changeAPDetailSignatureService.getApprovals(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, qualification);

      for each (var approvalListEntry in approvalList) {
        // this.log.debug("approvalListEntry [" + approvalListEntry + "]");

        // Set the approval object fields necessary to update a Remedy Approval
        setApproval(approvalListEntry, generateApprovalStatus(msg.response), responder, annotation);
        changeAPDetailSignatureService.setApprovalStatus(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, approvalListEntry);
        changeAPDetailSignatureService.setApproverSignature(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, approvalListEntry);
      }
      break;

    default:
      log.error("Unexpected response value: " + String(msg.response));
      break;
  }

  // Write to worklog
  setChangeWorkLog(change, true, NOTE_PREFIX + annotation, null);
  changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, change);

  log.debug("[Change ID " + infrastructureChangeId + "] " + annotation);
  log.debug("Exit - handleResponse");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleException
 *
 * Notify a responder about an error during the processing of the response
 *
 * @param msg message
 * @param e exception
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleException(msg, e) {
  log.debug("Enter - handleException");

  var infrastructureChangeId = getInfrastructureChangeId(msg);
  var errMsg = "Caught exception processing [" + infrastructureChangeId + "|RECEIVED_RESPONSE|" + msg.response + "]: " + e;
  log.error(errMsg);

  log.debug("Exit - handleException");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleEventStatus
 *
 * Handling event status messages from xMatters
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleEventStatus(msg) {
  log.debug("Enter - handleEventStatus");

  var eventStatusMessage = "";
  switch (msg.status) {
    case "active":
      // Event created in xMatters
      eventStatusMessage = "Event " + msg.eventidentifier + " successfully created in xMatters";
      break;

    case "terminated":
      // Event terminated
      eventStatusMessage = "Event " + msg.eventidentifier + " terminated";
      break;

    case "terminated_external": // event terminated by user if msg.username != INITIATOR
      if (msg.username === INITIATOR) {
        // Event terminated by this integration
        eventStatusMessage = "Event " + msg.eventidentifier + " terminated from within xMatters" + " by user [" + msg.username + "]";
      } else {
        // Event manually terminated from within xMatters
        eventStatusMessage = "Event " + msg.eventidentifier + " manually terminated from within xMatters" + " by user [" + msg.username + "]";
      }
      break;
  }
  if (eventStatusMessage.length > 0) {
    logNotification(eventStatusMessage);
  }

  log.debug("Exit - handleEventStatus");
}


