/**
 * ---------------------------------------------------------------------------------------------------------------------
 * apia_http
 *
 * This is the entry point for HTTP requests from Remedy that initiate data load operations.
 *
 * This method expects that the POST body contains a SOAP 1.1 message indicating the data load action to be taken.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function apia_http(httpRequestProperties, httpResponse) {
  log.info("Enter - apia_http");
  log.debug("httpRequestProperties: " + httpRequestProperties);

  var httpRequest = null;
  var responseBuilder = null;
  var responseBody = null;

  try {
    httpRequest = httpRequestProperties.getProperty("http.request");
    log.debug("Handling request to: " + httpRequest);

    var requestBodyWithoutXmlTag = httpRequestProperties.getProperty("REQUEST_BODY").replace("<?xml version='1.0' encoding='UTF-8'?>", "");
    processRequestBody(new XML(requestBodyWithoutXmlTag));

    responseBody = makeSoapResponseBody(APIA_HTTP_RECEIVED, "Request was submitted for processing");
  }
  catch (e) {
    log.warn("apia_http: caught Exception - name: [" + e.name + "], message [" + e.message + "]");
    responseBody = makeSoapResponseBody(APIA_HTTP_ERROR, e.toString());
  }
  finally {
    httpResponse.setStatusLine(HttpVersion.HTTP_1_1, 200);
    httpResponse.setHeader(new Header("Content-Type", httpRequestProperties.getProperty("Content-Type")));
    httpResponse.setBodyString(responseBody);
  }

  log.debug("apia_http - response body [" + responseBody + "]");
  log.info("Exit - apia_http");

  return httpResponse;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * processRequestBody
 *
 * @param requestBody - the body of the HTTP Request
 * ---------------------------------------------------------------------------------------------------------------------
 */
function processRequestBody(requestBody) {
  log.debug("Enter - processRequestBody");
  log.info("processRequestBody: request body [" + requestBody + "]");

  var request = new Object();

  request.action = String(requestBody. * ::Body. * ::TriggerRequest. * ::action);
  request.id = String(requestBody. * ::Body. * ::TriggerRequest. * ::id);
  request.requestType = String(requestBody. * ::Body. * ::TriggerRequest. * ::requestType);

  if (isEmpty(request.action)) {
    throw "Request action not found."
  }

  if (isEmpty(request.id)) {
    throw "Request id not found."
  }

  if (isEmpty(request.requestType)) {
    throw "Request type not found."
  }

  log.debug("Exit - processRequestBody");
  return processRequest(request);
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * processRequest
 *
 * Process the request from BMC Remedy, which has been unmarshalled into a object with properties for the:
 *
 * - request action (Delete or something else)
 * - request type (change or approval)
 * - id of the change or approval that the request applies to
 *
 * @param request -
 * ---------------------------------------------------------------------------------------------------------------------
 */
function processRequest(request) {
  log.info("Enter - processRequest");
  log.info("processRequestBody: action [" + request.action + "], requestType [" + request.requestType + "], id [" + request.id + "]");

  var xmWebService = new xMattersWS();

  // Check to see what kind of request action we are dealing with
  if (equalsIgnoreCase(request.action, REQUEST_ACTION_DELETE) && equalsIgnoreCase(request.requestType, REQUEST_TYPE_APPROVAL)) {
    // The request action is set to delete and the request type is approval
    // This means that an approval has moved to a finalized state (approved, rejected, closed, cancelled, error)
    var changeService = new XMCHGInfrastructureChangeWS(new WSUtil(), XM_CHG_INFRASTRUCTURE_CHANGE_WS_URL);

    // Retrieve the full description of the approval that the event is associated with, to get the Change Request
    var changeAPDetailSignatureService = new XMCHGChangeAPDetailSignatureWS(new WSUtil(), XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_URL);

    var qualification = XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_GET_APPROVAL_BY_REQUEST_ID_QUALIFICATION.replace("${REQUEST_ID}", String(request.id));

    var approvalList = changeAPDetailSignatureService.getApprovals(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, qualification);

    log.debug("processApprovalDel - approvalList for delete [" + approvalList + "]");

    var approvalID = approvalList[0].Approval_ID;
    var approvalStatus = approvalList[0].Approval_Status2;

    // We want to see if the approval status is Rejected
    // If the approval was rejected we want to delete any other events in xMatters which have been closed in Remedy
    // due to the Reject operation preformed in Remedy
    if (approvalStatus.equals("Rejected")) {

      log.debug("processApprovalDel - Sleep for " + TIME_TO_SLEEP_BEFORE_PROCESSING_DEL_FOR_REJECTED_APPOVALS + " ms to make sure the approvals status have been updated after the approval was rejected");
      try {
        Thread.sleep(TIME_TO_SLEEP_BEFORE_PROCESSING_DEL_FOR_REJECTED_APPOVALS);
      } catch (ex) {
        Thread.currentThread().interrupt();
      }

      // retrieve each approval within the current approval phase which has been marked as closed
      var qualification = XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_GET_CLOSED_APPROVALS_BY_APPROVAL_ID_QUALIFICATION.replace("${APPROVAL_ID}", String(approvalID));
      var closedApprovalsList = changeAPDetailSignatureService.getApprovals(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, qualification);

      if (closedApprovalsList != null && closedApprovalsList.length > 0) {
        log.info("processApprovalDel - closedApprovalsList: # to delete [" + closedApprovalsList.length + "] closedApprovalsList " + closedApprovalsList + "]");

        // for each closed approval send a DEL into xMatters to clean up the notifications
        for (var x = 0; x < closedApprovalsList.length; x++) {
          var approvalRequestID = closedApprovalsList[x].Request_ID;
          log.debug("processApprovalDel - deleting for reject approval request on approval [" + approvalID + "] with request id [" + approvalRequestID + "]");
          xmWebService.sendDelAPXML(approvalRequestID);
        }
      }

    }

    log.info("processRequest - deleting for approval request [" + request.id + "]");
    xmWebService.sendDelAPXML(request.id);
  }
  else if (equalsIgnoreCase(request.action, REQUEST_ACTION_DELETE)) {
    // Remedy can send requests to delete a change or approval request

    log.info("processRequest - deleting change or approval request [" + request.id + "]");
    xmWebService.sendDelAPXML(request.id);
  }
  else {
    // Request is a new or updated change or approval request

    var changeService = new XMCHGInfrastructureChangeWS(new WSUtil(), XM_CHG_INFRASTRUCTURE_CHANGE_WS_URL);
    var changeRequestId = "";

    // Even if we are processing an approval request notification, we need the ID of the Change Request.

    switch (String(request.requestType).toLowerCase()) {
      case REQUEST_TYPE_CHANGE:
        changeRequestId = request.id;
        break;

      case REQUEST_TYPE_APPROVAL:
        // Retrieve the full description of the approval that the event is associated with, to get the Change Request
        var changeAPDetailSignatureService = new XMCHGChangeAPDetailSignatureWS(new WSUtil(), XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_URL);
        var qualification = XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_GET_APPROVAL_BY_ID_QUALIFICATION.replace("${REQUEST_ID}", String(request.id));

        var approvalList = changeAPDetailSignatureService.getApprovals(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, qualification);

        log.debug("processApproval - approvalList [" + approvalList + "]");

        changeRequestId = approvalList[0].Infrastructure_Change_ID;
        break;

      default:
        throw ("Unknown request type [" + request.requestType + "]");
    }

    // Retrieve the Remedy Change Request associated with the event
    var changeRequest = changeService.getChangeRequest(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, changeRequestId);

    if (changeRequest == null) {
      throw ("changeRequest with id [" + request.id + "] not found in Remedy");
    }

    log.debug("processApproval - changeRequest [" + changeRequest + "]");

    // Make the APXML message to be sent to xMatters using the fields in the Change Request
    var apxml = makeApxmlFrom(changeRequest, request.id, String(request.requestType).toLowerCase());

    // Verify that the notification matches the appropriate CHANGE_COORDINATOR_ALERT_STATUS and CHANGE_MANAGER_ALERT_STATUS
    var status = String(apxml.getValue("change_request_status"));
    log.debug("The change request status is [" + status + "]");
    if (status != "" && CHANGE_COORDINATOR_ALERT_STATUS != null && CHANGE_COORDINATOR_ALERT_STATUS.indexOf(status) < 0) {
      // remove the coordinator login so a notification is not sent to that user
      apxml.setToken(CHANGE_COORDINATOR_LOGIN_FIELD, "");
      log.debug("The Change Coordinator will not be notified as the CHANGE_COORDINATOR_ALERT_STATUS list did not match the status of [" + status + "]");
    }
    if (status != "" && CHANGE_MANAGER_ALERT_STATUS != null && CHANGE_MANAGER_ALERT_STATUS.indexOf(status) < 0) {
      // remove the manager login so a notification is not sent to that user
      apxml.setToken(CHANGE_MANAGER_LOGIN_FIELD, "");
      log.debug("The Change Manager will not be notified as the CHANGE_MANAGER_ALERT_STATUS list did not match the status of [" + status + "]");
    }

    // If this is notification of an approval, add the tokens for the approval object
    if (String(request.requestType).toLowerCase() == REQUEST_TYPE_APPROVAL) {
      xmWebService.addEventTokensFromObject(apxml, approvalList[0]);
      apxml.setToken("apapprovers-list-item-delimiter", ";");

      // Replace the change request id with the change detail signature request id as the incident_id
      apxml.setToken("request_id", request.id);

      // Populate custom fields
      var jwt = REMEDYIO.generateRemedyApiToken();
      apxml.setToken("previous_approver", getPreviousApproverByRequestId(request.id, jwt));

      var approvalProcessName = String(apxml.getValue('approvalProcessName'));
      var apApprovers = String(apxml.getValue('APApprovers'));

      var alternateApprovers = "Lookup disabled";
      if (ALTERNATE_APPROVERS_LOOKUP) {
        // After getAlternateApprovers, alternateApprovers should hold the EXTRA approvers or blank if none.
        alternateApprovers = getAlternateApprovers(apApprovers, approvalProcessName, jwt);

        if (alternateApprovers.length > 0) {
          // Add a semi-colon if needed
          if (apApprovers.charAt(apApprovers.length - 1) !== ';') {
            apApprovers += ";";
          }

          apxml.setToken("target_recipients", apApprovers + alternateApprovers);
          apxml.setToken("alternate_approvers", alternateApprovers);
        } else {
          apxml.setToken("target_recipients", apApprovers);
          apxml.setToken("alternate_approvers", "None");
        }
      }
    }

    if (!SEND_ACTIVE_APPROVAL_ONLY ||
      (SEND_ACTIVE_APPROVAL_ONLY &&
        apxml.getValue("active_approval").equals("Yes") &&
        String(request.requestType).toLowerCase().equals(REQUEST_TYPE_APPROVAL))) {

      // Sending the payload to xMatters with xM API
      log.debug("Sending event to xMatters: URL=" + XMATTERS_REMEDY_CHANGE_APPROVAL_URL +
        " incident_id=" + apxml.getValue("incident_id") +
        " request_id=" + apxml.getValue("request_id"));

      // Add in response options.
      apxml.setToken("response_options_approve", RESPONSE_OPTIONS.APPROVE);
      apxml.setToken("response_options_reject", RESPONSE_OPTIONS.REJECT);
      apxml.setToken("response_options_hold", RESPONSE_OPTIONS.HOLD);
      apxml.setToken("response_options_view_change", RESPONSE_OPTIONS.VIEW_CHANGE);
      apxml.setToken("redirect_url", generateRedirectUrl(apxml.getValue("infrastructure_change_id")));

      var response = XMRESTAPI.submitApxml(XMATTERS_REMEDY_CHANGE_APPROVAL_URL, apxml,
        DELETE_EXISTING_EVENTS ? getExistingEventsFilter(apxml.getValue("incident_id")) : null, null);

      if (response) {
        log.debug("response.body: " + response.body);
        log.debug("response.status: " + response.status);

        // if request was actually sent
        if (response.status == XMRESTAPI.RESPONSE_ACCEPTED) {
          // Since IB doesn't return with an Event ID, use request id instead.
          log.info("Change Request " + request.id + " was sent to xMatters.");

          // process the event
          var changeInterfaceService = new XMCHGChangeInterfaceWS(new WSUtil(), XM_CHG_CHANGE_INTERFACE_WS_URL);

          setChangeWorkLog(changeRequest, true, NOTE_PREFIX + request.requestType + " event forwarded to xMatters", "");
          changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, changeRequest);

        } else {
          log.warn(request.id + ": FAILURE (" + response.status + ")");
          XMRESTAPI.checkResponse(response);
        }
      }
    } else {
      log.info('Not sending non-active-approval event.');
    }
  }

  log.info("Exit - processRequest");
  return apxml;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * makeApxmlFrom
 *
 * Constructs and returns an APXML message based on the information in the changeRequest and approvalList objects.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function makeApxmlFrom(changeRequest, id, requestType) {
  log.debug("Enter - makeApxmlFrom");
  log.debug("makeApxmlFrom - changeRequest [" + changeRequest + "], id [" + id + "], requestType [" + requestType + "]");

  var apxml = ServiceAPI.createAPXML();
  apxml.setMethod("ADD");
  apxml.setSubclass("action");

  // Set up the APXML tokens that have the same names and values as the Remedy changeRequest properties
  new xMattersWS().addEventTokensFromObject(apxml, changeRequest);

  apxml.setToken("incident_id", id);
  apxml.setToken("requestType", requestType);
  apxml.setToken("company_name", XMATTERS_COMPANY_NAME);

  // Add the xm_priority event token to set the priority of the event in xMatters (for 5.0)
  var xm_priority;

  switch (String(changeRequest.Priority).toLowerCase()) {
    case "critical" :
      xm_priority = "HIGH";
      break;
    case "high" :
      xm_priority = "MEDIUM";
      break;
    default :
      xm_priority = "LOW";
  }

  apxml.setToken("xm_priority", xm_priority);

  log.debug("Exit - makeApxmlFrom");
  return apxml;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * getExistingEventsFilter
 * @param incidentId ID of the incident ticket
 *
 * return a Javascript object that contains the event filter parameters
 * ---------------------------------------------------------------------------------------------------------------------
 */
function getExistingEventsFilter(incidentId) {
  return {properties: {'incident_id#en': incidentId}};
}

function getPreviousApproverByRequestId(requestId, jwt) {
  log.debug("Enter - getPreviousApproverByRequestId");
  var previousApprover = "N/A";

  var trimmedRequestedId = trimRequestId(requestId);
  var queryString = "%27Request%27%3d%22"+trimmedRequestedId+"%22AND%27Approval%20Status%27%3d%22Approved%22";
  var apiUrl = REMEDY_API_BASE_PATH + "AP:Detail-Signature?q=" + queryString;

  var response = REMEDYIO.get(apiUrl, "", "", REMEDYIO.generateHeaderForApiRequest(jwt));
  log.debug(response.status);
  log.debug(response.body);

  try {
    var body = JSON.parse(response.body);

    if (body.entries instanceof Array && body.entries.length > 0) {
      var latestApproval = body.entries[body.entries.length - 1].values;
      previousApprover = latestApproval['Approver Signature'];

    } else {
      log.info("Cannot get AP:Detail-Signature form data, or there is no previous approval for this request ID.");
    }
  } catch (e) {
    throw e;
  }

  log.debug("Exit - getPreviousApproverByRequestId");
  return previousApprover;
}

function getAlternateApprovers(approverNames, approvalProcessName, jwt) {
  log.debug("Enter - getAlternateApprovers");

  var alternateApprover = "";
  var originalApprovers = [];
  var alternateApprovers = [];

  // If there are approver names, populate the original array.
  if (approverNames) {
    var nameArray = approverNames.toString().split(";");
    for (var nameIdx in nameArray) {
      originalApprovers.push(nameArray[nameIdx]);
    }
  } else  {
    return alternateApprover;
  }

  // add to alternateApprovers array only if the Alternate isn't in the original array.
  try {
    originalApprovers.forEach(function(approver) {

      var queryString = "%27For%27%3d%22"+approver+"%22AND%27Status%27%3d%22Current%22";
      var apiUrl = REMEDY_API_BASE_PATH + "AP:Alternate?q=" + queryString;

      var response = REMEDYIO.get(apiUrl, "", "", REMEDYIO.generateHeaderForApiRequest(jwt));
      log.debug(response.status);
      log.debug(response.body);

      var body = JSON.parse(response.body);

      if (body.entries instanceof Array && body.entries.length > 0) {
        // Loop through the entires and check against originalApprovers;
        for (var i in body.entries) {
          var item = body.entries[i].values;
          var alternateFetched = item['Alternate'];
          var processNameFetched = item['Process Name'];

          if (originalApprovers.indexOf(alternateFetched) < 0 && alternateApprovers.indexOf(alternateFetched) < 0 &&
            (processNameFetched === null || processNameFetched.equals(approvalProcessName))) {
            log.debug("Adding alternate " + alternateFetched);
            alternateApprovers.push(alternateFetched);
          } else {
            log.debug("Not adding " + alternateFetched + ' to the list');
          }
        }
      } else {
        log.debug("Cannot get AP:Alternate form data, or there is no alternate for approver " + approver);
      }
    });

  } catch (e) {
    throw e;
  }

  log.debug("Exit - getAlternateApprovers");
  return alternateApprovers.join(";");
}



/**
 * requestId can com in two flavours:
 * - CRQnnnnnnnnnnnn
 * - CRQnnnnnnnnnnnn|nnnnnnnnnnnnnnn|nnnnnnnnnnnnnnn
 * Only the first section is returned.
 *
 * @param {string} requestId the request id
 * @returns {string} the first section of the requestId
 */
function trimRequestId(requestId) {
  if (!requestId) {
    throw "requestId is empty.";
  }

  if (requestId.indexOf("|") < 0) {
    return requestId;
  } else {
    return requestId.substr(0, requestId.indexOf("|"));
  }

}

/**
 * Generate a redirect url for View Change. It points to the change request ticket.
 * @param {string} infrastructureChangeId: the infrastructure change id
 * @returns {string} the url
 */
function generateRedirectUrl(infrastructureChangeId) {
  return "http://" + REMEDY_SERVER_NAME + ":" + MID_TIER_PORT + "/arsys/forms/" + REMEDY_SERVER_NAME +
    "/SHR%3ALandingConsole/Default+Administrator+View/?mode=search&F304255500=CHG%3AInfrastructure+Change" +
    "&F1000000076=FormOpen&F303647600=SearchTicketWithQual&F304255610='1000000182'%3D%22" + infrastructureChangeId+ "%22";
}
